/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Manage our cache of volume information.
 */

#include "arla_local.h"
RCSID("$Id: volcache.c,v 1.23 1998/02/23 04:05:43 assar Exp $") ;

/*
 * Suffixes for volume names.
 */

static char *volsuffixes[] = {
"",
ROSUFFIX,
BACKSUFFIX
};

/*
 * Return the root-volume.
 */

const char *
volcache_rootvolume (void)
{
     return "root.afs";		/* XXX */
}

#define VOLCACHE_SIZE 2048
#define VOLCACHE_INC 300

/* Hashtable of entries by name */
static Hashtab *volnamehashtab;

/* Hashtable of entries by number */
static Hashtab *volidhashtab;

/* A list with free entries */
static List *freelist;

/* # of entries */
static unsigned nvolcacheentries = 0;

/* # of active entries */
static unsigned nactive_volcacheentries = 0;

/*
 * VolCacheEntries are indexed by (name, cell) in volnamehashtab
 */

static int
volnamecmp (void *a, void *b)
{
    struct name_ptr *n1 = (struct name_ptr *)a;
    struct name_ptr *n2 = (struct name_ptr *)b;

    return strcmp (n1->name, n2->name)
	|| n1->cell != n2->cell;
}

static unsigned
volnamehash (void *a)
{
     struct name_ptr *n = (struct name_ptr *)a;

     return hashadd (n->name) + n->cell;
}

/*
 * and by (volid, cell) in volidhashtab
 */

static int
volidcmp (void *a, void *b)
{
    struct num_ptr *n1 = (struct num_ptr *)a;
    struct num_ptr *n2 = (struct num_ptr *)b;

    return n1->cell != n2->cell || n1->vol != n2->vol;
}

static unsigned
volidhash (void *a)
{
    struct num_ptr *n = (struct num_ptr *)a;

    return n->cell + n->vol;
}

/*
 * Create `n' entries and add to the `freelist'
 */

static void
create_new_entries (unsigned n)
{
    VolCacheEntry *entries;
    int i;

    entries = (VolCacheEntry *)calloc (n, sizeof(VolCacheEntry));
    if (entries == NULL)
	errx (1, "volcache: calloc failed");
    for (i = 0; i < n; ++i) {
	entries[i].cell = -1;
	listaddhead (freelist, &entries[i]);
    }
    
    nvolcacheentries += n;
}

/*
 * Initialize the volume cache with `nentries' in the free list.
 */

void
volcache_init (unsigned nentries)
{
    volnamehashtab = hashtabnew (VOLCACHE_SIZE, volnamecmp, volnamehash);
    if (volnamehashtab == NULL)
	errx (1, "volcache_init: hashtabnew failed");

    volidhashtab = hashtabnew (VOLCACHE_SIZE, volidcmp, volidhash);
    if (volidhashtab == NULL)
	errx (1, "volcache_init: hashtabnew failed");

    freelist = listnew ();
    if (freelist == NULL)
	errx (1, "volcache_init: listnew failed");
    nvolcacheentries = 0;
    create_new_entries (nentries);
}

/*
 * Re-cycle an entry:
 * remove it from the hashtab, clear it out and place it on the freelist.
 */

static void
recycle_entry (VolCacheEntry *e)
{
    int i;

    for (i = 0; i < MAXTYPES; ++i) {
	if (e->name_ptr[i].ptr != NULL)
	    hashtabdel (volnamehashtab, &e->name_ptr[i]);
	if (e->num_ptr[i].ptr != NULL)
	    hashtabdel (volidhashtab, &e->num_ptr[i]);
    }
    memset (e, 0, sizeof(*e));
    listaddhead (freelist, e);
    --nactive_volcacheentries;
}

/*
 * Get and return a free entry.
 */

static VolCacheEntry *
get_free_entry (void)
{
    VolCacheEntry *e;

    e = (VolCacheEntry *)listdelhead (freelist);
    if (e != NULL)
	return e;

    create_new_entries (VOLCACHE_INC);

    e = (VolCacheEntry *)listdelhead (freelist);
    assert (e != NULL);
    return e;
}

/*
 * Add an entry for (volname, cell) to the hash table.
 */

static VolCacheEntry *
add_entry (const char *volname, int32_t cell, CredCacheEntry *ce)
{
    VolCacheEntry *e;
    ConnCacheEntry *conn;
    u_long addr = cell_finddbserver (cell).s_addr;
    int error;

    e = get_free_entry ();

    e->cell = cell;
    e->refcount = 0;

    conn = conn_get (cell, addr, afsvldbport, VLDB_SERVICE_ID, ce);
    if (conn == NULL) {
	struct in_addr ia;

	listaddhead (freelist, e);
	ia.s_addr = addr;
	warnx ("volcache: failed to make rx-connection to vldb %s",
	       inet_ntoa(ia));
	return NULL;
    }

    error = VL_GetEntryByName(conn->connection, volname, &e->entry);
    conn_free (conn);
    if (error) {
	listaddhead (freelist, e);
	warnx ("volcache: VL_GetEntryByName(%s) failed: %d",
	       volname, error);
	return NULL;
    }

    if (e->entry.flags & VLF_RWEXISTS) {
	e->num_ptr[RWVOL].cell = cell;
	e->num_ptr[RWVOL].vol  = e->entry.volumeId[RWVOL];
	e->num_ptr[RWVOL].ptr  = e;
	hashtabadd (volidhashtab, (void *)&e->num_ptr[RWVOL]);

	e->name_ptr[RWVOL].cell = cell;
	snprintf (e->name_ptr[RWVOL].name, VLDB_MAXNAMELEN,
		  "%s%s", e->entry.name, volsuffixes[RWVOL]);
	e->name_ptr[RWVOL].ptr  = e;
	hashtabadd (volnamehashtab, (void *)&e->name_ptr[RWVOL]);

    }
    if (e->entry.flags & VLF_ROEXISTS) {
	e->num_ptr[ROVOL].cell = cell;
	e->num_ptr[ROVOL].vol  = e->entry.volumeId[ROVOL];
	e->num_ptr[ROVOL].ptr  = e;
	hashtabadd (volidhashtab, (void *)&e->num_ptr[ROVOL]);

	e->name_ptr[ROVOL].cell = cell;
	snprintf (e->name_ptr[ROVOL].name, VLDB_MAXNAMELEN,
		  "%s%s", e->entry.name, volsuffixes[ROVOL]);
	e->name_ptr[ROVOL].ptr  = e;
	hashtabadd (volnamehashtab, (void *)&e->name_ptr[ROVOL]);
    }
    if (e->entry.flags & VLF_BOEXISTS) {
	e->num_ptr[BACKVOL].cell = cell;
	e->num_ptr[BACKVOL].vol  = e->entry.volumeId[BACKVOL];
	e->num_ptr[BACKVOL].ptr  = e;
	hashtabadd (volidhashtab, (void *)&e->num_ptr[BACKVOL]);

	e->name_ptr[BACKVOL].cell = cell;
	snprintf (e->name_ptr[BACKVOL].name, VLDB_MAXNAMELEN,
		  "%s%s", e->entry.name, volsuffixes[BACKVOL]);
	e->name_ptr[BACKVOL].ptr  = e;
	hashtabadd (volnamehashtab, (void *)&e->name_ptr[BACKVOL]);
    }

    ++nactive_volcacheentries;

    return e;
}

/*
 * Retrieve the entry for (volname, cell).  If it's not in the cache,
 * add it.
 */

VolCacheEntry *
volcache_getbyname (const char *volname, int32_t cell, CredCacheEntry *ce)
{
    VolCacheEntry *e;
    struct name_ptr *n;
    struct name_ptr key;

    key.cell = cell;
    strncpy (key.name, volname, VLDB_MAXNAMELEN);
    key.name[VLDB_MAXNAMELEN - 1] = '\0';

    n = (struct name_ptr *)hashtabsearch (volnamehashtab, (void *)&key);
    if (n == NULL)
	e = add_entry (volname, cell, ce);
    else
	e = n->ptr;

    if (e != NULL)
	++e->refcount;
    return e;
}

/*
 * Retrieve the entry for (volume-id, cell). If it's not in the cache,
 * add it.
 */

VolCacheEntry *
volcache_getbyid (u_int32_t id, int32_t cell, CredCacheEntry *ce)
{
    VolCacheEntry *e;
    struct num_ptr *n;
    struct num_ptr key;

    key.cell = cell;
    key.vol  = id;

    n = (struct num_ptr *)hashtabsearch (volidhashtab, (void *)&key);
    if (n == NULL) {
	char s[11];

	snprintf (s, sizeof(s), "%u", id);
	e = add_entry (s, cell, ce);
    } else {
	e = n->ptr;
    }

    if (e != NULL)
	++e->refcount;

    return e;
}

/*
 * Save `volsync'
 */

void
volcache_update_volsync (VolCacheEntry *e, AFSVolSync volsync)
{
    e->volsync = volsync;
}


/*
 * Decrement the references and possibly remove this entry.
 */

void
volcache_free (VolCacheEntry *e)
{
    --e->refcount;
    if (e->refcount == 0)
	recycle_entry (e);
}

/*
 * Print the entry `ptr' to the FILE `arg'
 */

static Bool
print_entry (void *ptr, void *arg)
{
    struct num_ptr *n = (struct num_ptr *)ptr;
    VolCacheEntry *e = n->ptr;
    FILE *f = (FILE *)arg;
    int i;
    struct in_addr tmp;

    if (n->vol != e->entry.volumeId[RWVOL])
	return FALSE;

    fprintf (f, "cell = %d (%s)\n"
	     "name = \"%s\", volumeType = %d(%s), nServers = %d\n",
	     e->cell, cell_num2name (e->cell),
	     e->entry.name,
	     e->entry.volumeType, volsuffixes[e->entry.volumeType],
	     e->entry.nServers);
    for (i = 0; i < e->entry.nServers; ++i) {
	tmp.s_addr = htonl(e->entry.serverNumber[i]);
	fprintf (f, "%d: server = %s, part = %d(%c), flags = %d\n",
		 i, inet_ntoa(tmp), e->entry.serverPartition[i],
		 'a' + e->entry.serverPartition[i],
		 e->entry.serverFlags[i]);
    }
    if (e->entry.flags & VLF_RWEXISTS)
	fprintf (f, "rw clone: %d\n", e->entry.volumeId[RWVOL]);
    if (e->entry.flags & VLF_ROEXISTS)
	fprintf (f, "ro clone: %d\n", e->entry.volumeId[ROVOL]);
    if (e->entry.flags & VLF_BACKEXISTS)
	fprintf (f, "rw clone: %d\n", e->entry.volumeId[BACKVOL]);
    fprintf (f, "refcount = %u\n\n", e->refcount);
    return FALSE;
}

/*
 * Print some status on the volume cache on `f'.
 */

void
volcache_status (FILE *f)
{
    fprintf (f, "%u(%u) volume entries\n",
	     nactive_volcacheentries, nvolcacheentries);
    hashtabforeach (volidhashtab, print_entry, f);
}
