/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * XFS operations.
 */

#include <xfs_common.h>
#include <xfs_fs.h>
#include <xfs/xfs_node.h>
#include <xfs_message.h>
#include <xdeb.h>

static int
xfs_open_valid(struct vnode *vp, struct cred *cred, u_int tok)
{
  struct xfs *xfsp = XFS_FROM_VNODE(vp);
  struct xfs_node *xn = VNODE_TO_XNODE(vp);
  int errno = 0;
  
  do {
    if (!XFS_TOKEN_GOT(xn, tok))
      {
	struct xfs_message_open msg;
	msg.header.opcode = XFS_MSG_OPEN;
	msg.cred.uid = cred->cr_uid;
	msg.cred.pag = 0;	/* XXX */
	msg.handle = xn->handle;
	msg.tokens = tok;
	errno = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
	if (errno == 0)
	  errno = ((struct xfs_message_wakeup *) &msg)->errno;
      }
    else
      {
	goto done;
      }
  } while (errno == 0);

 done:
  return errno;
}

static int
xfs_attr_valid(struct vnode *vp, struct cred *cred, u_int tok)
{
  struct xfs *xfsp = XFS_FROM_VNODE(vp);
  struct xfs_node *xn = VNODE_TO_XNODE(vp);
  int errno = 0;
  
  do {
    if (!XFS_TOKEN_GOT(xn, tok))
      {
	struct xfs_message_getattr msg;
	msg.header.opcode = XFS_MSG_GETATTR;
	msg.cred.uid = cred->cr_uid;
	msg.cred.pag = 0;	/* XXX */
	msg.handle = xn->handle;
	errno = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
	if (errno == 0)
	  errno = ((struct xfs_message_wakeup *) &msg)->errno;
      }
    else
      {
	goto done;
      }
  } while (errno == 0);

 done:
  return errno;
}

static int
xfs_data_valid(struct vnode *vp, struct cred *cred, u_int tok)
{
  struct xfs *xfsp = XFS_FROM_VNODE(vp);
  struct xfs_node *xn = VNODE_TO_XNODE(vp);
  int errno = 0;
  
  do {
    if (!XFS_TOKEN_GOT(xn, tok))
      {
	struct xfs_message_getdata msg;
	msg.header.opcode = XFS_MSG_GETDATA;
	msg.cred.uid = cred->cr_uid;
	msg.cred.pag = 0;	/* XXX */
	msg.handle = xn->handle;
	errno = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
	if (errno == 0)
	  errno = ((struct xfs_message_wakeup *) &msg)->errno;
      }
    else
      {
	goto done;
      }
  } while (errno == 0);

 done:
  return errno;
}

static int
xfs_open(struct vnode **vpp,
	 int flag,
	 struct cred *cred)
{
  int errno = 0;
  
  XFSDEB(XDEBVNOPS, ("xfs_open\n"));
  
  if (flag & FWRITE)
    errno = xfs_open_valid(*vpp, cred, XFS_OPEN_NW);
  else
    errno = xfs_open_valid(*vpp, cred, XFS_OPEN_NR);
  
  return errno;
}

static int
xfs_close(struct vnode *vp,
	  int flag,
	  offset_t offset,
	  struct cred *cred)
{
  int errno = 0;

  XFSDEB(XDEBVNOPS, ("xfs_close\n"));
  
  if (flag & FWRITE)
    errno = xfs_fsync(vp, cred);
  
  return errno;
}

static int
xfs_read(struct vnode *vp,
	 struct uio *uio,
	 int ioflag,
	 struct cred *cred)
{
    int errno = 0;

    XFSDEB(XDEBVNOPS, ("xfs_read\n"));

    errno = xfs_data_valid(vp, cred, XFS_DATA_R);

    if (errno == 0) {
	struct vnode *t = DATA_FROM_VNODE(vp);
	VN_HOLD(t);
	errno = VOP_READ(t, uio, ioflag, cred);
	VN_RELE(t);
    }

    return errno;
}

static int
xfs_write(struct vnode *vp,
	  struct uio *uio,
	  int ioflag,
	  struct cred *cred)
{
    int errno = 0;

    XFSDEB(XDEBVNOPS, ("xfs_write\n"));

    errno = xfs_data_valid(vp, cred, XFS_DATA_W);

    if (errno == 0) {
	struct vnode *t = DATA_FROM_VNODE(vp);
	VN_HOLD(t);
	errno = VOP_WRITE(t, uio, ioflag, cred);
	VNODE_TO_XNODE(vp)->flags |= XFS_DATA_DIRTY;
	VN_RELE(t);
    }

    return errno;
}

static int
xfs_ioctl(struct vnode *vp,
	  int foo,
	  int bar,
	  struct cred *cred,
	  int *result)
{
  XFSDEB(XDEBVNOPS, ("xfs_ioctl\n"));
  return EINVAL;
}

static int
xfs_setfl(struct vnode *vp,
	  int foo,
	  int bar,
	  struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_setfl\n"));
  return EINVAL;
}

static int
xfs_getattr(struct vnode *vp,
	    struct vattr *vap,
	    int flags,
	    struct cred *cred)     
{
  int errno = 0;
  
  struct xfs_node *xn = VNODE_TO_XNODE(vp);

  XFSDEB(XDEBVNOPS, ("xfs_getattr\n"));
  
  errno = xfs_attr_valid(vp, cred, XFS_ATTR_R);
  if (errno == 0) {
      *vap = xn->attr;
  }
  
  return errno;
}

static int
xfs_setattr(struct vnode *vp,
	    struct vattr *vap,
	    int flags,
	    struct cred *cred)
{
  struct xfs *xfsp = XFS_FROM_VNODE(vp);
  struct xfs_node *xn = VNODE_TO_XNODE(vp);
  int errno = 0;
  
  XFSDEB(XDEBVNOPS, ("xfs_setattr\n"));
  if (XFS_TOKEN_GOT(xn, XFS_ATTR_W)) {
      /* Update attributes and mark them dirty. */
      VNODE_TO_XNODE(vp)->flags |= XFS_ATTR_DIRTY;
      errno = EINVAL;		/* XXX not yet implemented */
      goto done;
  } else {
      struct xfs_message_putattr msg;
      msg.header.opcode = XFS_MSG_PUTATTR;
      msg.cred.uid = cred->cr_uid;
      msg.cred.pag = 0;	/* XXX */
      msg.handle = xn->handle;
      vattr2xfs_attr (vap, &msg.attr);
      errno = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
      if (errno == 0)
	errno = ((struct xfs_message_wakeup *) &msg)->errno;
  }

 done:
  return errno;
}

static int
xfs_access(struct vnode *vp,
	   int mode,
	   int flags,
	   struct cred *cred)     
{
  int errno = 0;

  XFSDEB(XDEBVNOPS, ("xfs_access mode = 0%o\n", mode));

  errno = xfs_attr_valid(vp, cred, XFS_ATTR_R);
  if (errno == 0)
    {
      struct xfs_node *xn = VNODE_TO_XNODE(vp);
      mode >>= 6;		/* The kernel uses rwx------ */
      if (!((xn->id[0] == XFS_ANONYMOUSID)
	    && ((mode & ~xn->rights[0]) == 0)))
	{
	  int i;
	  errno = EACCES;	/* Until otherwise proven */
	  for (i = 0; i < MAXRIGHTS; i++)
	    if ((xn->id[i] == cred->cr_uid)
		&& (mode & ~xn->rights[i]))
	      {
		errno = 0;
		break;
	      }
	}   
    }

  XFSDEB(XDEBVNOPS, ("xfs_access(0%o) = %d\n", mode, errno));
  return 0;			/* XXX For now! */
  return errno;
}

static int
xfs_lookup(struct vnode *dvp,
	   char *nm,
	   struct vnode **vpp,
	   struct pathname *pnp,
	   int flags,
	   struct vnode *rdir,
	   struct cred *cred)
{
  struct xfs_message_getnode msg;
  struct xfs *xfsp = XFS_FROM_VNODE(dvp);
  int errno = 0;

  struct xfs_node *d = VNODE_TO_XNODE(dvp);
  struct vnode *v;
  
  XFSDEB(XDEBVNOPS, ("xfs_lookup\n"));
  
  do {
#ifdef notdef_but_correct
    errno = xfs_access(dvp, VEXEC, cred);
    if (errno != 0)
      goto done;
#endif
    v = xfs_dnlc_lookup(dvp, nm);
    if (!v)
      {
	msg.header.opcode = XFS_MSG_GETNODE;
	msg.cred.uid = cred->cr_uid;
	msg.cred.pag = 0;	/* XXX */
	msg.parent_handle = d->handle;
	strncpy(msg.name, nm, sizeof(msg.name));
	errno = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
	if (errno == 0)
	  errno = ((struct xfs_message_wakeup *) &msg)->errno;
      }
    else
      {
	*vpp = v;
	VN_HOLD(v);
	goto done;
      }
  } while (errno == 0);

 done:
  XFSDEB(XDEBVNOPS, ("xfs_lookup() = %d\n", errno));
  return errno;
}

static int
xfs_create(struct vnode *dvp,
	   char *nm,
	   struct vattr *va,
	   vcexcl_t exclusive,
	   int mode,
	   struct vnode **vpp,
	   struct cred *cred)     
{
  struct xfs *xfsp = XFS_FROM_VNODE(dvp);
  struct xfs_node *xn = VNODE_TO_XNODE(dvp);
  int errno = 0;
  
  XFSDEB(XDEBVNOPS, ("xfs_create\n"));
  {
    struct xfs_message_create msg;
    msg.header.opcode = XFS_MSG_CREATE;
    msg.parent_handle = xn->handle;
    strncpy(msg.name, nm, 256);
    vattr2xfs_attr (va, &msg.attr);
#if 0
    msg.exclusive = exclusive;
#endif
    msg.mode = mode;
    msg.cred.uid = cred->cr_uid;
    msg.cred.pag = 0;	/* XXX */
    errno = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (errno == 0)
      errno = ((struct xfs_message_wakeup *) &msg)->errno;
  }

  if (errno == 0)
    errno = xfs_lookup(dvp, nm, vpp, /*pnp*/ NULL, /*flags*/ 0,
		       /*rdir*/ NULL, cred);
  return errno;
}

static int
xfs_remove(struct vnode *dvp,
	   char *nm,
	   struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_remove\n"));
  return EINVAL;
}

static int
xfs_link(struct vnode *vp,
	 struct vnode *tdvp,
	 char *tnm,
	 struct cred *cred)     
{
  XFSDEB(XDEBVNOPS, ("xfs_link\n"));
  return EINVAL;
}

static int
xfs_rename(struct vnode *odvp,
	   char *onm,
	   struct vnode *ndvp,
	   char *nnm,
	   struct cred *cred)     
{
  XFSDEB(XDEBVNOPS, ("xfs_rename\n"));
  return EINVAL;
}

static int
xfs_mkdir(struct vnode *dvp,
	  char *nm,
	  struct vattr *va,
	  struct vnode **vpp,
	  struct cred *cred)     
{
  struct xfs *xfsp = XFS_FROM_VNODE(dvp);
  struct xfs_node *xn = VNODE_TO_XNODE(dvp);
  int errno = 0;
  
  XFSDEB(XDEBVNOPS, ("xfs_mkdir\n"));
  {
    struct xfs_message_mkdir msg;
    msg.header.opcode = XFS_MSG_CREATE;
    msg.parent_handle = xn->handle;
    strncpy(msg.name, nm, 256);
    vattr2xfs_attr (va, &msg.attr);
    msg.cred.uid = cred->cr_uid;
    msg.cred.pag = 0;	/* XXX */
    errno = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (errno == 0)
      errno = ((struct xfs_message_wakeup *) &msg)->errno;
  }

  return errno;
}

static int
xfs_rmdir(struct vnode *dvp,
	  char *nm,
	  struct vnode *foo,
	  struct cred *cred)     
{
  XFSDEB(XDEBVNOPS, ("xfs_rmdir\n"));
  return EINVAL;
}

static int
xfs_readdir(struct vnode *vp,
	    struct uio *uiop,
	    struct cred *cred,
	    int *eofp)		/* XXX */
{
  int errno = 0;

  XFSDEB(XDEBVNOPS, ("xfs_readdir\n"));

  errno = xfs_data_valid(vp, cred, XFS_DATA_R);
  if (errno == 0)
    {
      struct vnode *t = DATA_FROM_VNODE(vp);
      VN_HOLD(t);
      errno = VOP_READ(t, uiop, 0, cred);
      VN_RELE(t);
    }

  return errno;
}

static int
xfs_symlink(struct vnode *dvp,
	    char *lnm,
	    struct vattr *tva,
	    char *tnm,
	    struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_symlink\n"));
  return EINVAL;
}

static int
xfs_readlink(struct vnode *vp,
	     struct uio *uiop,
	     struct cred *cred)
{
  int errno = 0;

  XFSDEB(XDEBVNOPS, ("xfs_readlink\n"));

  errno = xfs_data_valid(vp, cred, XFS_DATA_R);
  if (errno == 0)
    {
      struct vnode *t = DATA_FROM_VNODE(vp);
      VN_HOLD(t);
      errno = VOP_READ(t, uiop, 0, cred);
      VN_RELE(t);
    }

  return errno;
}

static int
xfs_fsync(struct vnode *vp,
	  struct cred *cred)
{
  struct xfs *xfsp = XFS_FROM_VNODE(vp);
  struct xfs_node *xn = VNODE_TO_XNODE(vp);
  int errno = 0;
  
  XFSDEB(XDEBVNOPS, ("xfs_fsync\n"));
  if (VNODE_TO_XNODE(vp)->flags & XFS_DATA_DIRTY)
    {
      struct xfs_message_putdata msg;
      msg.header.opcode = XFS_MSG_PUTDATA;
      msg.cred.uid = cred->cr_uid;
      msg.cred.pag = 0;	/* XXX */
      msg.handle = xn->handle;
      errno = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
      if (errno == 0)
	errno = ((struct xfs_message_wakeup *) &msg)->errno;
    }
  
  return errno;
}

static int
xfs_inactive(struct vnode *vp,
	     struct cred *cred)
{
  struct xfs_message_inactivenode msg;
  struct xfs *xfsp = XFS_FROM_VNODE(vp);
  struct xfs_node *xn = VNODE_TO_XNODE(vp);

  XFSDEB(XDEBVNOPS, ("xfs_inactive\n"));

  msg.header.opcode = XFS_MSG_INACTIVENODE;
  msg.handle = xn->handle;
  free_xfs_node(VNODE_TO_XNODE(vp));
  xfs_message_send(xfsp->fd, &msg.header, sizeof(msg));
  return 0;
}

static int
xfs_fid(struct vnode *vp,
	struct fid *fid)
{
  XFSDEB(XDEBVNOPS, ("xfs_fid\n"));
  return EINVAL;
}

static int
xfs_rwlock(struct vnode *vp,
	   int foo)
{
  XFSDEB(XDEBVNOPS, ("xfs_rwlock\n"));
  return EINVAL;
}

static int
xfs_rwunlock(struct vnode *vp,
	     int foo)
{
  XFSDEB(XDEBVNOPS, ("xfs_rwunlock\n"));
  return EINVAL;
}

static int
xfs_seek(struct vnode *vp,
	 offset_t offset,
	 offset_t *roffset)
{
  XFSDEB(XDEBVNOPS, ("xfs_seek\n"));
  return EINVAL;
}

static int
xfs_cmp(struct vnode *vp1, struct vnode *vp2)
{
  XFSDEB(XDEBVNOPS, ("xfs_cmp\n"));
  return EINVAL;
}

static int
xfs_frlock(struct vnode *vp,
	   int foo,
	   struct flock *fl,
	   int bar,
	   offset_t off,
	   struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_frlock\n"));
  return EINVAL;
}

static int
xfs_space(struct vnode *vp,
	  int foo,
	  struct flock *fl,
	  int bar,
	  offset_t offset,
	  struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_space\n"));
  return EINVAL;
}

static int
xfs_realvp(struct vnode *vp,
	   struct vnode **vpp)
{
  XFSDEB(XDEBVNOPS, ("xfs_realvp\n"));
  return EINVAL;
}

static int
xfs_getpage(struct vnode *vp,
	    offset_t off,
	    u_int len,
	    u_int *protp,
	    struct page *pl[],
	    u_int plsz,
	    struct seg *seg,
	    caddr_t addr,
	    enum seg_rw rw,
	    struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_getpage\n"));
  return EINVAL;
}

static int
xfs_putpage(struct vnode *vp,
	    offset_t off,
	    u_int len,
	    int flags,
	    struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_putpage\n"));
  return EINVAL;
}

static int
xfs_map(struct vnode *vp,
	offset_t off,
	struct as *as,
	caddr_t *addrp,
	u_int len,
	u_char prot,
	u_char maxprot,
	u_int flags,
	struct cred *cred)
{
  int errno = 0;

  XFSDEB(XDEBVNOPS,
	 ("xfs_map(0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x)\n",
	  (int) vp, off, (int) as, (int) addrp, len, prot, maxprot, flags, (int) cred));

  if ((prot & PROT_WRITE) && (flags & MAP_SHARED))
    errno = xfs_data_valid(vp, cred, XFS_DATA_W);
  else
    errno = xfs_data_valid(vp, cred, XFS_DATA_R);
  
  if (errno != 0)
    /* Can't map today */;
  else if (off + len > VNODE_TO_XNODE(vp)->attr.va_size)
    errno = EINVAL;
  else if ((prot & PROT_WRITE) && (flags & MAP_SHARED))
    errno = EROFS;		/* XXX This is currently not supported */
  else
    {
      struct vnode *t = DATA_FROM_VNODE(vp);
      VN_HOLD(t);
      errno = VOP_MAP(t, off, as, addrp, len, prot, maxprot, flags, cred);
      /* XXX Patch vnode so that we can intercept get/putpage and inactive. */
      VN_RELE(t);
    }

  return errno;
}

static int
xfs_addmap(struct vnode *vp,
	   offset_t off,
	   struct as *as,
	   caddr_t *addrp,
	   u_int len,
	   u_char prot,
	   u_char maxprot,
	   u_int flags,
	   struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_addmap\n"));
  return EINVAL;

}

static int
xfs_delmap(struct vnode *vp,
	   offset_t off,
	   struct as *as,
	   caddr_t *addrp,
	   u_int len,
	   u_char prot,
	   u_char maxprot,
	   u_int flags,
	   struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_delmap\n"));
  return EINVAL;

}


static int
xfs_poll(struct vnode *vp,
	 short events,
	 int anyyet,
	 short *revents,
	 struct pollhead **ph)
{
  XFSDEB(XDEBVNOPS, ("xfs_poll\n"));
  return EINVAL;
}

static int
xfs_dump(struct vnode *dumpvp,
	 caddr_t addr,
	 int bn,
	 int count)
{
  XFSDEB(XDEBVNOPS, ("xfs_dump\n"));
  return EINVAL;
}

static int
xfs_pathconf(struct vnode *vp,
	     int foo,
	     u_long *bar,
	     struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_pathconf\n"));
  return EINVAL;
}

static int
xfs_pageio(struct vnode *vp,
	   struct page *page,
	   u_int a,
	   u_int b,
	   int c,
	   struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_pageio\n"));
  return EINVAL;
}

static int
xfs_dumpctl(struct vnode *vp,
	    int flag)
{
  XFSDEB(XDEBVNOPS, ("xfs_dumpctl\n"));
  return EINVAL;
}

static int
xfs_dispose(struct vnode *vp,
	    struct page *page,
	    int a,
	    int b,
	    struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_dispose\n"));
  return EINVAL;
}

static int
xfs_setsecattr(struct vnode *vp,
	       vsecattr_t *attr,
	       int flag,
	       struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_setsecattr\n"));
  return EINVAL;
}

static int
xfs_getsecattr(struct vnode *vp,
	       vsecattr_t *attr,
	       int flag,
	       struct cred *cred)
{
  XFSDEB(XDEBVNOPS, ("xfs_getsecattr\n"));
  return EINVAL;
}

struct vnodeops xfs_vnodeops = {
        xfs_open,
        xfs_close,
        xfs_read,
        xfs_write,
        xfs_ioctl,
	xfs_setfl,
        xfs_getattr,
        xfs_setattr,
        xfs_access,
        xfs_lookup,
        xfs_create,
        xfs_remove,
        xfs_link,
        xfs_rename,
        xfs_mkdir,
        xfs_rmdir,
        xfs_readdir,
        xfs_symlink,
        xfs_readlink,
        xfs_fsync,
        xfs_inactive,
	xfs_fid,
	xfs_rwlock,
	xfs_rwunlock,
	xfs_seek,
	xfs_cmp,
	xfs_frlock,
	xfs_space,
	xfs_realvp,
        xfs_getpage,
        xfs_putpage,
        xfs_map,
	xfs_addmap,
	xfs_delmap,
	xfs_poll,
        xfs_dump,
	xfs_pathconf,
	xfs_pageio,
	xfs_dumpctl,
	xfs_dispose,
	xfs_setsecattr,
	xfs_getsecattr
};
